@@ -3,7 +3,7 @@
  *
  * Additional autoconf defines for this program.
  *
- * $Id: acconfig.h,v 1.1.1.1 1999/12/17 16:30:14 patl Exp $
+ * $Id: acconfig.h,v 1.4 2004/04/22 10:48:16 quozl Exp $
  */
 
 /* Use BSD User land PPP? */
@@ -15,6 +15,9 @@
 /* Let PPPD choose the IP addresses? */
 #undef PPPD_IP_ALLOC
 
+/* Enable Broadcast Relay? */
+#undef BCRELAY
+
 /* Work as a PNS rather than a PAC? */
 #undef PNS_MODE
 
