@@ -3,16 +3,18 @@
  *
  * Compatibility functions for different OSes
  *
- * $Id: compat.c,v 1.1.1.1 1999/12/17 16:30:14 patl Exp $
+ * $Id: compat.c,v 1.4 2004/04/22 10:48:16 quozl Exp $
  */
 
 #if HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#ifndef HAVE_STRLCPY
 #include "compat.h"
+
+#ifndef HAVE_STRLCPY
 #include <string.h>
+#include <stdio.h>
 
 void strlcpy(char *dst, const char *src, size_t size)
 {
@@ -100,3 +102,25 @@ char *strerror(int errnum) {
 	return buf;
 }
 #endif
+
+#ifndef HAVE_SETPROCTITLE
+#include "inststr.h"
+#endif
+
+#define __USE_BSD 1
+#include <stdarg.h>
+#include <stdio.h>
+
+void my_setproctitle(int argc, char **argv, const char *format, ...) {
+       char proctitle[64];
+       va_list parms;
+       va_start(parms, format);
+       vsnprintf(proctitle, sizeof(proctitle), format, parms);
+
+#ifndef HAVE_SETPROCTITLE
+       inststr(argc, argv, proctitle);
+#else
+       setproctitle(proctitle);
+#endif
+       va_end(parms);
+}
