@@ -3,7 +3,7 @@
  *
  * Compatibility functions for different OSes (prototypes)
  *
- * $Id: compat.h,v 1.1.1.1 1999/12/17 16:30:14 patl Exp $
+ * $Id: compat.h,v 1.4 2004/04/22 10:48:16 quozl Exp $
  */
 
 #ifndef _PPTPD_COMPAT_H
@@ -13,6 +13,12 @@
 #include "config.h"
 #endif
 
+#if HAVE_SETSID
+#define SETSIDPGRP setsid
+#else
+#define SETSIDPGRP setpgrp
+#endif
+
 #include <sys/types.h>
 
 #ifndef HAVE_STRLCPY
@@ -51,4 +57,7 @@ extern int openpty(int *, int *, char *, void *, void *);
 extern char *strerror(int);
 #endif
 
+extern void my_setproctitle(int argc, char **argv, const char *format, ...)
+       __attribute__ ((format (printf, 3, 4)));
+
 #endif	/* !_PPTPD_COMPAT_H */
