@@ -3,7 +3,7 @@
  *
  * PPTP Control Message packet reading, formatting and writing.
  *
- * $Id: ctrlpacket.c,v 1.3 1999/12/23 21:43:33 patl Exp $
+ * $Id: ctrlpacket.c,v 1.5 2004/04/29 06:58:45 quozl Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -20,6 +20,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/types.h>
+#include <time.h>
 #include <sys/time.h>
 #include <netinet/in.h>
 #include <unistd.h>
@@ -213,6 +214,7 @@ ssize_t read_pptp_header(int clientFd, unsigned char *packet, int *pptp_ctrl_typ
 	if (buffered) {
 		memcpy(packet, buffer, buffered);
 		free(buffer);
+		buffer = NULL;
 		bytes_ttl = buffered;
 		buffered = 0;
 		if (pptpctrl_debug)
@@ -230,6 +232,8 @@ ssize_t read_pptp_header(int clientFd, unsigned char *packet, int *pptp_ctrl_typ
 				if (bytes_ttl) {
 					buffered = bytes_ttl;
 					buffer = malloc(bytes_ttl);
+					if (!buffer)
+						return(0);
 					memcpy(buffer, packet, bytes_ttl);
 				}
 				syslog(LOG_ERR, "CTRL: Error reading ctrl packet length (bytes_ttl=%d): %s", bytes_ttl, strerror(errno));
@@ -245,6 +249,8 @@ ssize_t read_pptp_header(int clientFd, unsigned char *packet, int *pptp_ctrl_typ
 			if (bytes_ttl == 1) {
 				buffered = bytes_ttl;
 				buffer = malloc(bytes_ttl);
+				if (!buffer)
+					return(0);
 				memcpy(buffer, packet, bytes_ttl);
 				if (pptpctrl_debug)
 					syslog(LOG_DEBUG, "CTRL: Incomplete ctrl packet length, retry later");
@@ -252,10 +258,19 @@ ssize_t read_pptp_header(int clientFd, unsigned char *packet, int *pptp_ctrl_typ
 			}
 		}
 	}
-	/* OK, we have (at least) the first 2 bytes, and there is data waiting */
+	/* OK, we have (at least) the first 2 bytes, and there is data waiting
+	 *
+	 * length includes the header,  so a length less than 2 is someone
+	 * trying to hack into us or a badly corrupted packet.
+	 * Why not require length to be at least 10? Since we later cast
+	 * packet to struct pptp_header and use at least the 10 first bytes..
+	 * Thanks to Timo Sirainen for mentioning this.
+	 */
 	length = htons(*(u_int16_t *) packet);
-	if (length > PPTP_MAX_CTRL_PCKT_SIZE) {
-		syslog(LOG_ERR, "CTRL: Control packet > PPTP_MAX_CTRL_PCKT_SIZE (length = %d)", length);
+	if (length <= 10 || length > PPTP_MAX_CTRL_PCKT_SIZE) {
+		syslog(LOG_ERR, "CTRL: 11 < Control packet (length=%d) < "
+				"PPTP_MAX_CTRL_PCKT_SIZE (%d)",
+				length, PPTP_MAX_CTRL_PCKT_SIZE);
 		/* we loose sync (unless we malloc something big, which isn't a good
 		 * idea - potential DoS) so we must close connection (draft states that
 		 * if you loose sync you must close the control connection immediately)
@@ -271,6 +286,8 @@ ssize_t read_pptp_header(int clientFd, unsigned char *packet, int *pptp_ctrl_typ
 			if (bytes_ttl) {
 				buffered = bytes_ttl;
 				buffer = malloc(bytes_ttl);
+				if (!buffer)
+					return(0);
 				memcpy(buffer, packet, bytes_ttl);
 			}
 			syslog(LOG_ERR, "CTRL: Error reading ctrl packet (bytes_ttl=%d,length=%d): %s", bytes_ttl, length, strerror(errno));
@@ -286,6 +303,8 @@ ssize_t read_pptp_header(int clientFd, unsigned char *packet, int *pptp_ctrl_typ
 		if (bytes_ttl != length) {
 			buffered = bytes_ttl;
 			buffer = malloc(bytes_ttl);
+			if (!buffer)
+				return(0);
 			memcpy(buffer, packet, bytes_ttl);
 			if (pptpctrl_debug)
 				syslog(LOG_DEBUG, "CTRL: Incomplete ctrl packet, retry later");
@@ -347,9 +366,6 @@ void deal_start_ctrl_conn(unsigned char *packet, unsigned char *rply_packet, ssi
 	struct pptp_start_ctrl_conn_rqst *start_ctrl_conn_rqst;
 	struct pptp_start_ctrl_conn_rply start_ctrl_conn_rply;
 
-	unsigned char pptpd_hostname[MAX_HOSTNAME_SIZE] = PPTP_HOSTNAME;
-	unsigned char pptpd_vendor[MAX_VENDOR_SIZE] = PPTP_VENDOR;
-
 	start_ctrl_conn_rqst = (struct pptp_start_ctrl_conn_rqst *) packet;
 
 	MAKE_CTRL_HEADER(start_ctrl_conn_rply, START_CTRL_CONN_RPLY);
@@ -360,8 +376,10 @@ void deal_start_ctrl_conn(unsigned char *packet, unsigned char *rply_packet, ssi
 	start_ctrl_conn_rply.bearer_cap = htons(OUR_BEARER);
 	start_ctrl_conn_rply.max_channels = htons(MAX_CHANNELS);
 	start_ctrl_conn_rply.firmware_rev = htons(PPTP_FIRMWARE_VERSION);
-	memcpy(start_ctrl_conn_rply.hostname, pptpd_hostname, MAX_HOSTNAME_SIZE);
-	memcpy(start_ctrl_conn_rply.vendor, pptpd_vendor, MAX_VENDOR_SIZE);
+	bzero(start_ctrl_conn_rply.hostname, MAX_HOSTNAME_SIZE);
+	strncpy(start_ctrl_conn_rply.hostname, PPTP_HOSTNAME, MAX_HOSTNAME_SIZE);
+	bzero(start_ctrl_conn_rply.vendor, MAX_VENDOR_SIZE);
+	strncpy(start_ctrl_conn_rply.vendor, PPTP_VENDOR, MAX_VENDOR_SIZE);
 	COPY_CTRL_PACKET(start_ctrl_conn_rply, rply_packet, rply_size);
 	DEBUG_PACKET("START CTRL CONN RPLY");
 }
@@ -423,12 +441,12 @@ void deal_out_call(unsigned char *packet, unsigned char *rply_packet, ssize_t *
 	 * still performs at over 115200, eg, 60kbyte/sec and higher observed.
 	 */
 	out_call_rply.speed = out_call_rqst->max_bps;
-	out_call_rply.pckt_recv_size = hton16(PCKT_RECV_WINDOW_SIZE);
+	/* lets match their window size for now... was htons(PCKT_RECV_WINDOW_SIZE)
+	 */
+	out_call_rply.pckt_recv_size = out_call_rqst->pckt_recv_size;
 	if(pptpctrl_debug)
-		syslog(LOG_DEBUG, "CTRL: %d min_bps, %d max_bps, %d window size",
-                       ntohs(out_call_rqst->min_bps),
-                       ntohs(out_call_rqst->max_bps),
-                       ntohs(out_call_rply.pckt_recv_size));
+		syslog(LOG_DEBUG, "CTRL: Set parameters to %d maxbps, %d window size",
+			ntohl(out_call_rply.speed), ntohs(out_call_rply.pckt_recv_size));
 	out_call_rply.pckt_delay = htons(PCKT_PROCESS_DELAY);
 	out_call_rply.channel_id = htonl(CHANNEL_ID);
 	COPY_CTRL_PACKET(out_call_rply, rply_packet, rply_size);
@@ -630,6 +648,25 @@ static u_int16_t _pac_init = 0;
 u_int16_t getcall()
 {
 	static u_int16_t i = 0;
+	extern u_int16_t unique_call_id;
+
+	/* Start with a random Call ID.  This is to allocate unique
+	 * Call ID's across multiple TCP PPTP connections.  In this
+	 * way remote clients masqueraded by a firewall will put
+	 * unique peer call ID's into GRE packets that will have the
+	 * same source IP address of the firewall. */
+
+	if (!i) {
+		if (unique_call_id == 0xFFFF) {
+			struct timeval tv;
+			if (gettimeofday(&tv, NULL) == 0) {
+				i = ((tv.tv_sec & 0x0FFF) << 4) + 
+				    (tv.tv_usec >> 16);
+			}
+		} else {
+			i = unique_call_id;
+		}
+	}
 
 	if(!_pac_init) {
 		_pac_call_id = htons(-1);
@@ -637,8 +674,9 @@ u_int16_t getcall()
 	}
 	if(_pac_call_id != htons(-1))
 		syslog(LOG_ERR, "CTRL: Asked to allocate call id when call open, not handled well");
-	_pac_call_id = i;
-	return i++;
+	_pac_call_id = htons(i);
+	i++;
+	return _pac_call_id;
 }
 
 /*
