@@ -3,7 +3,7 @@
  *
  * PPTP control connection between PAC-PNS pair
  *
- * $Id: pptpctrl.c,v 1.4 1999/12/29 16:06:41 patl Exp $
+ * $Id: pptpctrl.c,v 1.13 2004/04/28 11:36:07 quozl Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -14,11 +14,7 @@
 #define _GNU_SOURCE 1		/* kill() prototype, broken arpa/inet.h */
 #endif
 
-#if HAVE_SYSLOG_H
-#include <syslog.h>
-#else
 #include "our_syslog.h"
-#endif
 
 #include <fcntl.h>
 #include <errno.h>
@@ -27,8 +23,11 @@
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <time.h>
 #include <sys/time.h>
+#include <dirent.h>
 #include <sys/types.h>
+#include <sys/wait.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -41,19 +40,25 @@
 #endif
 #endif
 
-#ifndef HAVE_SETPROCTITLE
-#include "inststr.h"
+#ifdef __UCLIBC__
+#define socklen_t int
 #endif
+
 #include "compat.h"
 #include "pptpctrl.h"
 #include "pptpgre.h"
 #include "pptpdefs.h"
 #include "ctrlpacket.h"
 #include "defaults.h"
+// placing net/if.h here fixes build on Solaris
+#include <net/if.h>
 
-/* Globals because i'm lazy -tmk */
+static char *ppp_binary = PPP_BINARY;
+static int pptp_logwtmp;
+static int noipparam;			/* if true, don't send ipparam to ppp */
 static char speed[32];
 static char pppdxfig[256];
+static pid_t pppfork;                   /* so we can kill it after disconnect */
 
 /*
  * Global to handle dying
@@ -66,12 +71,17 @@ static u_int32_t call_id_pair;	/* call id (to terminate call) */
 
 /* Needed by this and ctrlpacket.c */
 int pptpctrl_debug = 0;		/* specifies if debugging is on or off */
+uint16_t unique_call_id = 0xFFFF;	/* Start value for our call IDs on this TCP link */
+
+int gargc;                     /* Command line argument count */
+char **gargv;                  /* Command line argument vector */
 
 /* Local function prototypes */
 static void bail(int sigraised);
 static void pptp_handle_ctrl_connection(char **pppaddrs, struct in_addr *inetaddrs);
-static int startCall(char **pppaddrs);
-static void launch_pppd(char **pppaddrs);
+
+static int startCall(char **pppaddrs, struct in_addr *inetaddrs);
+static void launch_pppd(char **pppaddrs, struct in_addr *inetaddrs);
 
 /* Oh the horror.. lets hope this covers all the ones we have to handle */
 #if defined(O_NONBLOCK) && !defined(__sun__) && !defined(__sun)
@@ -80,57 +90,52 @@ static void launch_pppd(char **pppaddrs);
 #define OUR_NB_MODE O_NDELAY
 #endif
 
-/* Simple function */
-static void die_if_invalid(char *ptr) {
-	if (!ptr) {
-		fprintf(stderr,
-"pptpctrl: Ran out of command line arguments.\n\n"
-"\tpptpctrl should either be run from pptpd or from inetd as instructed\n"
-"\tin pptpd(8) or pptpctrl(8) manual pages.\n"
-			);
-		exit(2);
-	}
-}
-
 /* Macro to read command line args */
 #define GETARG(X) \
-	die_if_invalid(*argv); \
-	if(atoi(argv[arg++]) != 0) { \
-		die_if_invalid(*argv); \
+	if(atoi(argv[arg++]) != 0) \
 		strlcpy(X, argv[arg++], sizeof(X)); \
-	} else \
+	else \
 		*X = '\0'
 
 int main(int argc, char **argv)
 {
 	char pppLocal[16];		/* local IP to pass to pppd */
 	char pppRemote[16];		/* remote IP address to pass to pppd */
-	char ipx[16];			/* IPX network number */
 	struct sockaddr_in addr;	/* client address */
 	socklen_t addrlen;
 	int arg = 1;
-#ifndef HAVE_SETPROCTITLE
-	char proctitle[32];
-#endif
 	int flags;
 	struct in_addr inetaddrs[2];
-	char *pppaddrs[3] = { pppLocal, pppRemote, ipx };
+	char *pppaddrs[2] = { pppLocal, pppRemote };
+
+        gargc = argc;
+        gargv = argv;
+
+	/* fail if argument count invalid */
+	if (argc < 7) {
+		fprintf(stderr, "pptpctrl: insufficient arguments, see man pptpctrl\n");
+		exit(2);
+	}
 
 	/* open a connection to the syslog daemon */
-	openlog("pptpd", LOG_PID, LOG_PPTP);
+	openlog("pptpd", LOG_PID, LOG_DAEMON);
 
 	/* autoreap if supported */
 	signal(SIGCHLD, SIG_IGN);
 
-	die_if_invalid(*argv);
+	/* note: update pptpctrl.8 if the argument list format is changed */
 	pptpctrl_debug = atoi(argv[arg++]);
+	noipparam = atoi(argv[arg++]);
 
 	GETARG(pppdxfig);
 	GETARG(speed);
 	GETARG(pppLocal);
 	GETARG(pppRemote);
-	GETARG(ipx);
 
+	if (arg < argc) unique_call_id = atoi(argv[arg++]);
+	if (arg < argc) ppp_binary = strdup(argv[arg++]);
+	if (arg < argc) pptp_logwtmp = atoi(argv[arg++]);
+	
 	if (pptpctrl_debug) {
 		if (*pppLocal)
 			syslog(LOG_DEBUG, "CTRL: local address = %s", pppLocal);
@@ -145,7 +150,7 @@ int main(int argc, char **argv)
 	addrlen = sizeof(addr);
 	if (getsockname(clientSocket, (struct sockaddr *) &addr, &addrlen) != 0) {
 		syslog(LOG_ERR, "CTRL: getsockname() failed");
-		perror("getsockname");
+		syslog_perror("getsockname");
 		close(clientSocket);
 		bail(0);	/* NORETURN */
 	}
@@ -154,7 +159,7 @@ int main(int argc, char **argv)
 	addrlen = sizeof(addr);
 	if (getpeername(clientSocket, (struct sockaddr *) &addr, &addrlen) != 0) {
 		syslog(LOG_ERR, "CTRL: getpeername() failed");
-		perror("getpeername");
+		syslog_perror("getpeername");
 		close(clientSocket);
 		bail(0);	/* NORETURN */
 	}
@@ -164,19 +169,15 @@ int main(int argc, char **argv)
 	if ((flags = fcntl(clientSocket, F_GETFL, arg /* ignored */)) == -1 ||
 	    fcntl(clientSocket, F_SETFL, flags|OUR_NB_MODE) == -1) {
 		syslog(LOG_ERR, "CTRL: Failed to set client socket non-blocking");
-		perror("fcntl");
+		syslog_perror("fcntl");
 		close(clientSocket);
 		bail(0);	/* NORETURN */
 	}
 
 	
 	/* Fiddle with argv */
-#if HAVE_SETPROCTITLE
-	setproctitle("pptpd [%s]", inet_ntoa(addr.sin_addr));
-#else
-	sprintf(proctitle, "pptpd [%s]", inet_ntoa(addr.sin_addr));
-	inststr(argc, argv, proctitle);
-#endif
+        my_setproctitle(gargc, gargv, "pptpd [%s]%20c",
+            inet_ntoa(addr.sin_addr), ' ');
 
 	/* be ready for a grisly death */
 	signal(SIGTERM, &bail);
@@ -185,6 +186,9 @@ int main(int argc, char **argv)
 
 	syslog(LOG_INFO, "CTRL: Client %s control connection started", inet_ntoa(addr.sin_addr));
 	pptp_handle_ctrl_connection(pppaddrs, inetaddrs);
+	syslog(LOG_DEBUG, "CTRL: Reaping child PPP[%i]", pppfork);
+	if (pppfork > 0)
+		waitpid(pppfork, NULL, 0);
 	syslog(LOG_INFO, "CTRL: Client %s control connection finished", inet_ntoa(addr.sin_addr));
 
 	bail(0);		/* NORETURN */
@@ -209,233 +213,227 @@ int main(int argc, char **argv)
  */
 static void pptp_handle_ctrl_connection(char **pppaddrs, struct in_addr *inetaddrs)
 {
-  /* For echo requests used to check link is alive */
-  int echo_wait = FALSE;		/* Waiting for echo? */
-  u_int32_t echo_count = 0;	/* Sequence # of echo */
-  time_t echo_time;		/* Time last echo req sent */
-  struct timeval idleTime;	/* How long to select() */
 
-  /* General local variables */
-  ssize_t rply_size;		/* Reply packet size */
-  fd_set readfds;		/* For select() */
-  fd_set writefds;              /* For select() */
-  int maxfd = clientSocket;	/* For select() */
-  int send_packet;		/* Send a packet this time? */
+	/* For echo requests used to check link is alive */
+	int echo_wait = FALSE;		/* Waiting for echo? */
+	u_int32_t echo_count = 0;	/* Sequence # of echo */
+	time_t echo_time = 0;		/* Time last echo req sent */
+	struct timeval idleTime;	/* How long to select() */
+
+	/* General local variables */
+	ssize_t rply_size;		/* Reply packet size */
+	fd_set fds;			/* For select() */
+	int maxfd = clientSocket;	/* For select() */
+	int send_packet;		/* Send a packet this time? */
 #if BSDUSER_PPP || SLIRP
 /* not needed by stuff which uses socketpair() in startCall() */
 #define init 1
 #else
-  int init = 0;			/* Has pppd initialized the pty? */
+	int init = 0;			/* Has pppd initialized the pty? */
 #endif
-  struct gre_state *gre = NULL;
+	int pty_fd = -1;		/* File descriptor of pty */
+	int gre_fd = -1;		/* Network file descriptor */
 
-  unsigned char packet[PPTP_MAX_CTRL_PCKT_SIZE];
-  unsigned char rply_packet[PPTP_MAX_CTRL_PCKT_SIZE];
-  u_int16_t max_window_size;
-  int new_pty_fd = -1;
+	unsigned char packet[PPTP_MAX_CTRL_PCKT_SIZE];
+	unsigned char rply_packet[PPTP_MAX_CTRL_PCKT_SIZE];
 
-  for (;;) {
-    FD_ZERO(&readfds);
-    FD_ZERO(&writefds);
+	for (;;) {
 
-    /* always willing to read from the control channel */
-    FD_SET (clientSocket, &readfds);
+		FD_ZERO(&fds);
+		FD_SET(clientSocket, &fds);
+		if (pty_fd != -1)
+			FD_SET(pty_fd, &fds);
+		if (gre_fd != -1 && init)
+			FD_SET(gre_fd, &fds);
 
-    /* always willing to read from the GRE socket */
-    if (gre && gre->gre_fd != -1 && init)
-      FD_SET (gre->gre_fd, &readfds);
+		/* set timeout */
+		if (encaps_gre(-1, NULL, 0) || decaps_hdlc(-1, NULL, 0)) {
+			idleTime.tv_sec = 0;
+			idleTime.tv_usec = 50000; /* don't ack immediately */
+		} else {
+			idleTime.tv_sec = IDLE_WAIT;
+			idleTime.tv_usec = 0;
+		}
 
-    /* Willing to read from the PTY if the buffer is empty */
-    if (gre && gre->pty_fd != -1
-        && gre->pty_bytes_processed == gre->pty_bytes_read)
-      FD_SET (gre->pty_fd, &readfds);
+		/* default: do nothing */
+		send_packet = FALSE;
 
-    /* want to write to the GRE socket if we have a packet ready */
-    if (gre && gre->gre_fd != -1 && gre->gre_packet_ready)
-      FD_SET (gre->gre_fd, &writefds);
+		switch (select(maxfd + 1, &fds, NULL, NULL, &idleTime)) {
+		case -1:	/* Error with select() */
+			if (errno != EINTR)
+				syslog(LOG_ERR, "CTRL: Error with select(), quitting");
+			goto leave_clear_call;
 
-    /* want to write to the PTY if the output buffer is nonempty */
-    if (gre && gre->pty_fd != -1 && init
-        && gre->pty_output_size != gre->pty_bytes_written)
-      FD_SET (gre->pty_fd, &writefds);
+		case 0:
+			if (decaps_hdlc(-1, NULL, 0)) {
+				if(decaps_hdlc(-1, encaps_gre, gre_fd))
+					syslog(LOG_ERR, "CTRL: GRE re-xmit failed");
+			} else if (encaps_gre(-1, NULL, 0))
+				/* Pending ack and nothing else to do */
+				encaps_gre(gre_fd, NULL, 0);	/* send ack with no payload */
+			else if (echo_wait != TRUE) {
+				/* Timeout. Start idle link detection. */
+				echo_count++;
+				if (pptpctrl_debug)
+					syslog(LOG_DEBUG, "CTRL: Sending ECHO REQ id %d", echo_count);
+				time(&echo_time);
+				make_echo_req_packet(rply_packet, &rply_size, echo_count);
+				echo_wait = TRUE;
+				send_packet = TRUE;
+			}
+			break;
 
-    /* set timeout */
-    /* FIXME: need to do adaptive timeouts / windowing */
-    idleTime.tv_sec = IDLE_WAIT;
-    idleTime.tv_usec = 0;
+		default:
+			break;
+		}
 
-    /* default: do nothing */
-    send_packet = FALSE;
-
-    switch (select(maxfd + 1, &readfds, &writefds, NULL, &idleTime)) {
-    case -1:	/* Error with select() */
-      if (errno != EINTR)
-        syslog(LOG_ERR, "CTRL: Error with select(), quitting");
-      goto leave_clear_call;
-
-    case 0:
-      if (echo_wait != TRUE) {
-        /* Timeout. Start idle link detection. */
-        echo_count++;
-        if (pptpctrl_debug)
-          syslog(LOG_DEBUG, "CTRL: Sending ECHO REQ id %d", echo_count);
-        time(&echo_time);
-        make_echo_req_packet(rply_packet, &rply_size, echo_count);
-        echo_wait = TRUE;
-        send_packet = TRUE;
-      }
-      break;
-
-    default:
-      break;
-    }
-
-    /* detect startup of pppd */
+		/* detect startup of pppd */
 #ifndef init
-    if (!init && gre && gre->pty_fd != -1 && FD_ISSET(gre->pty_fd, &readfds))
-      init = 1;
+		if (!init && pty_fd != -1 && FD_ISSET(pty_fd, &fds))
+			init = 1;
 #endif
 
-    /* handle actual packets */
+		/* handle actual packets */
 
-    /* send from pty off via GRE */
-    if (gre) {
-      if (do_gre_to_pty (gre) < 0) {
-        syslog(LOG_ERR,
-               "CTRL: PTY read or GRE write failed (pty,gre)=(%d,%d)",
-               gre->pty_fd, gre->gre_fd);
-        break;
-      }
+		/* send from pty off via GRE */
+		if (pty_fd != -1 && FD_ISSET(pty_fd, &fds) && decaps_hdlc(pty_fd, encaps_gre, gre_fd) < 0) {
+			syslog(LOG_ERR, "CTRL: PTY read or GRE write failed (pty,gre)=(%d,%d)", pty_fd, gre_fd);
+			break;
+		}
+		/* send from GRE off to pty */
+		if (gre_fd != -1 && FD_ISSET(gre_fd, &fds) && decaps_gre(gre_fd, encaps_hdlc, pty_fd) < 0) {
+			if (gre_fd == 6 && pty_fd == 5) {
+				syslog(LOG_ERR, "CTRL: GRE-tunnel has collapsed (GRE read or PTY write failed (gre,pty)=(%d,%d))", gre_fd, pty_fd);
+			} else {
+				syslog(LOG_ERR, "CTRL: GRE read or PTY write failed (gre,pty)=(%d,%d)", gre_fd, pty_fd);
+			}
+			break;
+		}
+		/* handle control messages */
 
-      if (do_pty_to_gre (gre) < 0) {
-        syslog(LOG_ERR,
-               "CTRL: GRE read or PTY write failed (gre,pty)=(%d,%d)",
-               gre->gre_fd, gre->pty_fd);
-        break;
-      }
-    }
+		if (FD_ISSET(clientSocket, &fds)) {
+			send_packet = TRUE;
+			switch (read_pptp_packet(clientSocket, packet, rply_packet, &rply_size)) {
+			case 0:
+				syslog(LOG_ERR, "CTRL: CTRL read failed");
+				goto leave_drop_call;
 
-    if (FD_ISSET(clientSocket, &readfds)) {
-      send_packet = TRUE;
-      switch (read_pptp_packet(clientSocket, packet,
-                               rply_packet, &rply_size)) {
-      case 0:
-        syslog(LOG_ERR, "CTRL: CTRL read failed");
-        goto leave_drop_call;
+			case -1:
+				send_packet = FALSE;
+				break;
 
-      case -1:
-        send_packet = FALSE;
-        break;
+			case STOP_CTRL_CONN_RQST:
+				if (pptpctrl_debug)
+					syslog(LOG_DEBUG, "CTRL: Received STOP CTRL CONN request (disconnecting)");
+				if (gre_fd != -1 || pty_fd != -1)
+					syslog(LOG_WARNING, "CTRL: Request to close control connection when call is open, closing");
+				send_pptp_packet(clientSocket, rply_packet, rply_size);
+				goto leave_drop_call;
 
-      case STOP_CTRL_CONN_RQST:
-        if (pptpctrl_debug)
-          syslog(LOG_DEBUG, "CTRL: Received STOP CTRL CONN request (disconnecting)");
-        if (gre)
-          syslog(LOG_WARNING, "CTRL: closing...");
-        send_pptp_packet(clientSocket, rply_packet, rply_size);
-        goto leave_drop_call;
+			case CALL_CLR_RQST:
+				if(pptpctrl_debug)
+					syslog(LOG_DEBUG, "CTRL: Received CALL CLR request (closing call)");
+				if (gre_fd == -1 || pty_fd == -1)
+					syslog(LOG_WARNING, "CTRL: Request to close call but call not open");
+				if (gre_fd != -1) {
+					FD_CLR(gre_fd, &fds);
+					close(gre_fd);
+					gre_fd = -1;
+				}
+				if (pty_fd != -1) {
+					FD_CLR(pty_fd, &fds);
+					close(pty_fd);
+					pty_fd = -1;
+				}
+				/* violating RFC */
+                                goto leave_drop_call;
 
-      case CALL_CLR_RQST:
-        if(pptpctrl_debug)
-          syslog(LOG_DEBUG, "CTRL: Received CALL CLR request (closing call)");
-        if (!gre || gre->gre_fd == -1 || gre->pty_fd == -1)
-          syslog(LOG_WARNING, "CTRL: Request to close call but call not open");
-        if (gre) {
-          if (gre->gre_fd != -1) {
-            close(gre->gre_fd);
-            gre->gre_fd = -1;
-          }
-          if (gre->pty_fd != -1) {
-            close(gre->pty_fd);
-            gre->pty_fd = -1;
-          }
-          NOTE_VALUE(PAC, call_id_pair, htons(-1));
-          NOTE_VALUE(PNS, call_id_pair, htons(-1));
-          break;
-        }
+			case OUT_CALL_RQST:
+				/* for killing off the link later (ugly) */
+				NOTE_VALUE(PAC, call_id_pair, ((struct pptp_out_call_rply *) (rply_packet))->call_id);
+				NOTE_VALUE(PNS, call_id_pair, ((struct pptp_out_call_rply *) (rply_packet))->call_id_peer);
+				if (gre_fd != -1 || pty_fd != -1) {
+					syslog(LOG_WARNING, "CTRL: Request to open call when call is already open, closing");
+					if (gre_fd != -1) {
+						FD_CLR(gre_fd, &fds);
+						close(gre_fd);
+						gre_fd = -1;
+					}
+					if (pty_fd != -1) {
+						FD_CLR(pty_fd, &fds);
+						close(pty_fd);
+						pty_fd = -1;
+					}
+				}
+				/* Start the call */
+				syslog(LOG_INFO, "CTRL: Starting call (launching pppd, opening GRE)");
 
-      case OUT_CALL_RQST:
-        /* for killing off the link later (ugly) */
-        NOTE_VALUE (PAC, call_id_pair,
-                    ((struct pptp_out_call_rply *) (rply_packet))->call_id);
-        NOTE_VALUE(PNS, call_id_pair, ((struct pptp_out_call_rply *)
-                                       (rply_packet))->call_id_peer);
-        max_window_size =
-          ntoh16 (((struct pptp_out_call_rqst *)packet)->pckt_recv_size);
-        if (gre) {
-          syslog(LOG_WARNING, "CTRL: Request to open call when call is already open, closing");
-          if (gre->gre_fd != -1) {
-            close (gre->gre_fd);
-            gre->gre_fd = -1;
-          }
-          if (gre->pty_fd != -1) {
-            close (gre->pty_fd);
-            gre->pty_fd = -1;
-          }
-          free (gre);
-        }
-        /* Start the call */
-        syslog(LOG_INFO, "CTRL: Starting call (launching pppd, opening GRE)");
-        if ((new_pty_fd = startCall(pppaddrs)) > maxfd)
-          maxfd = new_pty_fd;
-        gre = pptp_gre_init(call_id_pair, max_window_size,
-                            new_pty_fd, inetaddrs);
-        if (!gre) {
-          syslog (LOG_ERR, "pptp_gre_init failed; exiting");
-          exit (-1);
-        }
+                                /* Fiddle with argv */
+                                my_setproctitle(gargc, gargv,
+                                      "pptpd [%s:%04X - %04X]",
+                                      inet_ntoa(inetaddrs[1]),
+                                      ntohs(((struct pptp_out_call_rply *) (rply_packet))->call_id_peer),
+                                      ntohs(((struct pptp_out_call_rply *) (rply_packet))->call_id));
 
-        if (gre->gre_fd > maxfd)
-          maxfd = gre->gre_fd;
-        break;
+				if ((pty_fd = startCall(pppaddrs, inetaddrs)) > maxfd)
+					maxfd = pty_fd;
+				if ((gre_fd = pptp_gre_init(call_id_pair, pty_fd, inetaddrs)) > maxfd)
+					maxfd = gre_fd;
+				break;
 
-      case ECHO_RPLY:
-        if (echo_wait == TRUE && ((struct pptp_echo_rply *) (packet))->identifier == echo_count)
-          echo_wait = FALSE;
-        else
-          syslog(LOG_WARNING, "CTRL: Unexpected ECHO REPLY packet");
-        /* FALLTHRU */
-      case SET_LINK_INFO:
-        send_packet = FALSE;
-        break;
+			case ECHO_RPLY:
+				if (echo_wait == TRUE && ((struct pptp_echo_rply *) (packet))->identifier == echo_count)
+					echo_wait = FALSE;
+				else
+					syslog(LOG_WARNING, "CTRL: Unexpected ECHO REPLY packet");
+				/* FALLTHRU */
+			case SET_LINK_INFO:
+				send_packet = FALSE;
+				break;
 
-      case CALL_DISCONN_NTFY:
-      case STOP_CTRL_CONN_RPLY:
-        /* These don't generate replies.  Also they come from things
-           we don't send in this section. */
-        syslog(LOG_WARNING, "CTRL: Got a reply to a packet we didn't send");
-        send_packet = FALSE;
-        break;
+#ifdef PNS_MODE
+			case IN_CALL_RQST:
+			case IN_CALL_RPLY:
+			case IN_CALL_CONN:
+#endif
 
-        /* Otherwise, the already-formed reply will do fine, so send it */
-      }
-    }
+			case CALL_DISCONN_NTFY:
+			case STOP_CTRL_CONN_RPLY:
+				/* These don't generate replies.  Also they come from things we don't send in this section. */
+				syslog(LOG_WARNING, "CTRL: Got a reply to a packet we didn't send");
+				send_packet = FALSE;
+				break;
 
-    /* send reply packet - this may block, but it should be very rare */
-    if (send_packet == TRUE
-        && send_pptp_packet(clientSocket, rply_packet, rply_size) < 0) {
-      syslog(LOG_ERR, "CTRL: Error sending GRE, aborting call");
-      goto leave_clear_call;
-    }
+			/* Otherwise, the already-formed reply will do fine, so send it */
+			}
+		}
 
-    /* waiting for echo reply and curtime - echo_time > max wait */
-    if (echo_wait == TRUE && (time(NULL) - echo_time) > MAX_ECHO_WAIT) {
-      syslog(LOG_INFO, "CTRL: Session timed out, ending call");
-      goto leave_clear_call;
-    }
-  }
-  /* Finished! :-) */
+		/* send reply packet - this may block, but it should be very rare */
+		if (send_packet == TRUE && send_pptp_packet(clientSocket, rply_packet, rply_size) < 0) {
+			syslog(LOG_ERR, "CTRL: Error sending GRE, aborting call");
+			goto leave_clear_call;
+		}
+
+		/* waiting for echo reply and curtime - echo_time > max wait */
+		if (echo_wait == TRUE && (time(NULL) - echo_time) > MAX_ECHO_WAIT) {
+			syslog(LOG_INFO, "CTRL: Session timed out, ending call");
+			goto leave_clear_call;
+		}
+	}
+	/* Finished! :-) */
 leave_drop_call:
-  NOTE_VALUE(PAC, call_id_pair, htons(-1));
-  NOTE_VALUE(PNS, call_id_pair, htons(-1));
-  close(clientSocket);
+	NOTE_VALUE(PAC, call_id_pair, htons(-1));
+	NOTE_VALUE(PNS, call_id_pair, htons(-1));
+	close(clientSocket);
 leave_clear_call:
-  /* leave clientSocket and call_id_pair alone for bail() */
-  if (gre->gre_fd != -1)
-    close(gre->gre_fd);
-  if (gre->pty_fd != -1)
-    close(gre->pty_fd);
-  return;
+	/* leave clientSocket and call_id_pair alone for bail() */
+	if (gre_fd != -1)
+		close(gre_fd);
+	gre_fd = -1;
+	if (pty_fd != -1)
+		close(pty_fd);
+	pty_fd = -1;
+	return;
 #ifdef init
 #undef init
 #endif
@@ -464,6 +462,7 @@ static void bail(int sigraised)
 		unsigned char rply_packet[PPTP_MAX_CTRL_PCKT_SIZE];
 		ssize_t rply_size;	/* reply packet size */
 		int pkt;
+		int retry = 0;
 
 		if (pptpctrl_debug)
 			syslog(LOG_DEBUG, "CTRL: Exiting with active call");
@@ -482,17 +481,9 @@ static void bail(int sigraised)
 		tv.tv_usec = 0;
 
 		/* Wait for STOP CTRL CONN RQST or RPLY */
-		while (select(clientSocket + 1, &connSet, NULL, NULL,
-                              &tv) == 1) {
-                  pkt = read_pptp_packet(clientSocket, packet,
-                                         rply_packet, &rply_size);
-                  if (pkt == 0) {
-                    syslog (LOG_WARNING,
-                            "EOF reading control message");
-                    break;
-                  }
-                  else switch (pkt) {
-                        case STOP_CTRL_CONN_RQST:
+		while (select(clientSocket + 1, &connSet, NULL, NULL, &tv) == 1) {
+			switch((pkt = read_pptp_packet(clientSocket, packet, rply_packet, &rply_size))) {
+			case STOP_CTRL_CONN_RQST:
 				send_pptp_packet(clientSocket, rply_packet, rply_size);
 				goto skip;
 			case CALL_CLR_RQST:
@@ -500,26 +491,32 @@ static void bail(int sigraised)
 				break;
 			case STOP_CTRL_CONN_RPLY:
 				goto skip;
-			case 0:
-				syslog(LOG_WARNING, "CTRL: Client already disconnected control stream");
-				goto skip;
 			case -1:
-				syslog(LOG_WARNING, "CTRL: Error reading control stream during disconnect, aborting");
+				syslog(LOG_WARNING, "CTRL: Retryable error in disconnect sequence");
+				retry++;
+				break;
+			case 0:
+				syslog(LOG_WARNING, "CTRL: Fatal error reading control message in disconnect sequence");
 				goto skip;
 			default:
 				syslog(LOG_WARNING, "CTRL: Unexpected control message %d in disconnect sequence", pkt);
+				retry++;
+				break;
 			}
 			tv.tv_sec = 5;	/* wait 5 secs for another reply then quit */
 			tv.tv_usec = 0;
+			if (retry > 100) {
+				syslog(LOG_WARNING, "CTRL: Too many retries (%d) - giving up", retry);
+				break;
+			}
 		}
-skip:
+
+	skip:
 		close(clientSocket);
 	}
 
 	if (pptpctrl_debug)
 		syslog(LOG_DEBUG, "CTRL: Exiting now");
-
-	exit((GET_VALUE(PAC, call_id_pair)==htons(-1)) ? 0 : 1);
 }
 
 /*
@@ -531,11 +528,11 @@ skip:
  * retn:        pty file descriptor
  *
  */
-static int startCall(char **pppaddrs)
+static int startCall(char **pppaddrs, struct in_addr *inetaddrs)
 {
 	/* PTY/TTY pair for talking to PPPd */
 	int pty_fd, tty_fd;
-        int flags;
+	/* register pids of children */
 #if BSDUSER_PPP || SLIRP
 	int sockfd[2];
 
@@ -550,8 +547,8 @@ static int startCall(char **pppaddrs)
 	/* userspace ppp doesn't need to waste a real pty/tty pair */
 	if (socketpair(AF_LOCAL, SOCK_STREAM, 0, sockfd)) {
 		syslog(LOG_ERR, "CTRL: socketpair() error");
-		perror("socketpair");
-		exit(3);
+		syslog_perror("socketpair");
+		exit(1);
 	}
 	tty_fd = sockfd[0];
 	pty_fd = sockfd[1];
@@ -559,8 +556,8 @@ static int startCall(char **pppaddrs)
 	/* Finds an open pty/tty pair */
 	if (openpty(&pty_fd, &tty_fd, NULL, NULL, NULL) != 0) {
 		syslog(LOG_ERR, "CTRL: openpty() error");
-		perror("openpty");
-		exit(3);
+		syslog_perror("openpty");
+		exit(1);
 	}
 #endif
 	if (pptpctrl_debug) {
@@ -569,13 +566,13 @@ static int startCall(char **pppaddrs)
 	}
 	/* Launch the PPPD  */
 #ifndef HAVE_FORK
-	switch (vfork()) {
+        switch(pppfork=vfork()){
 #else
-	switch (fork()) {
+        switch(pppfork=fork()){
 #endif
 	case -1:	/* fork() error */
 		syslog(LOG_ERR, "CTRL: Error forking to exec pppd");
-		break;
+		_exit(1);
 
 	case 0:		/* child */
 		dup2(tty_fd, 0);
@@ -587,49 +584,51 @@ static int startCall(char **pppaddrs)
 		 */
 		dup2(tty_fd, 2);
 #endif
-		if (tty_fd > 2)
+		if (tty_fd > 1)
 			close(tty_fd);
-		if (pty_fd > 2)
+		if (pty_fd > 1)
 			close(pty_fd);
 /* In case we move clientSocket back off stdin */
 #ifndef clientSocket
-		if (clientSocket > 2)
+		if (clientSocket > 1)
 			close(clientSocket);
-#elif clientSocket > 2
+#elif clientSocket > 1
 		close(clientSocket);
 #endif
-		launch_pppd(pppaddrs);
-		syslog(LOG_ERR, "CTRL: PPPD launch failed!");
-		exit(3);
+		launch_pppd(pppaddrs, inetaddrs);
+		syslog(LOG_ERR, "CTRL: PPPD launch failed! (launch_pppd did not fork)");
+		_exit(1);
 	}
+	
 	close(tty_fd);
-
-        flags = fcntl (pty_fd, F_GETFL, 0);
-        if (flags < 0)
-          syslog (LOG_ERR, "CTRL: Couldn't get flags for TTY!");
-        else if (fcntl (pty_fd, F_SETFL, flags | O_NONBLOCK) < 0)
-          syslog (LOG_ERR, "CTRL: Cound't set flags for TTY!");
-
 	return pty_fd;
 }
 
-
 /*
- * lanuch_pppd
+ * launch_pppd
  *
  * Launches the PPP daemon. The PPP daemon is responsible for assigning the
  * PPTP client its IP address.. These values are assigned via the command
  * line.
  *
+ * Add return of connected ppp interface
+ *
  * retn: 0 on success, -1 on failure.
  *
  */
-static void launch_pppd(char **pppaddrs)
+static void launch_pppd(char **pppaddrs, struct in_addr *inetaddrs)
 {
-	char *pppd_argv[16];
+	char *pppd_argv[14];
 	int an = 0;
+	sigset_t sigs;
 
-	pppd_argv[an++] = PPP_BINARY;
+	pppd_argv[an++] = ppp_binary;
+
+	if (pptpctrl_debug) {
+		syslog(LOG_DEBUG, 
+		       "CTRL (PPPD Launcher): program binary = %s", 
+		       pppd_argv[an - 1]);
+	}
 
 #if BSDUSER_PPP
 
@@ -694,6 +693,7 @@ static void launch_pppd(char **pppaddrs)
 #else
 
 	/* options for 'normal' pppd */
+
 	pppd_argv[an++] = "local";
 
 	/* If a pppd option file is specified, use it
@@ -703,6 +703,7 @@ static void launch_pppd(char **pppaddrs)
 		pppd_argv[an++] = "file";
 		pppd_argv[an++] = pppdxfig;
 	}
+	
 	/* If a speed has been specified, use it
 	 * if not, use "smart" default (defaults.h)
 	 */
@@ -713,29 +714,41 @@ static void launch_pppd(char **pppaddrs)
 	}
 
 	if (pptpctrl_debug) {
-		syslog(LOG_DEBUG, "CTRL (PPPD Launcher): Connection speed = %s", pppd_argv[an - 1]);
 		if (*pppaddrs[0])
 			syslog(LOG_DEBUG, "CTRL (PPPD Launcher): local address = %s", pppaddrs[0]);
 		if (*pppaddrs[1])
 			syslog(LOG_DEBUG, "CTRL (PPPD Launcher): remote address = %s", pppaddrs[1]);
-		if (*pppaddrs[2])
-			syslog(LOG_DEBUG, "CTRL (PPPD Launcher): ipx network = %s", pppaddrs[2]);
 	}
+	
 	if (*pppaddrs[0] || *pppaddrs[1]) {
 		char pppInterfaceIPs[33];
 		sprintf(pppInterfaceIPs, "%s:%s", pppaddrs[0], pppaddrs[1]);
 		pppd_argv[an++] = pppInterfaceIPs;
 	}
-
-	if (*pppaddrs[2]) {
-		pppd_argv[an++] = "ipx-network";
-		pppd_argv[an++] = pppaddrs[2];
-	}
 #endif
 
+        if (!noipparam) {
+                 pppd_argv[an++] = "ipparam";
+                 pppd_argv[an++] = inet_ntoa(inetaddrs[1]);
+        }
+
+        if (pptp_logwtmp) {
+                 pppd_argv[an++] = "plugin";
+                 pppd_argv[an++] = "/usr/lib/pptpd/pptpd-logwtmp.so";
+                 pppd_argv[an++] = "pptpd-original-ip";
+                 pppd_argv[an++] = inet_ntoa(inetaddrs[1]);
+        }
+
 	/* argv arrays must always be NULL terminated */
-	pppd_argv[an] = NULL;
+	pppd_argv[an++] = NULL;
+	/* make sure SIGCHLD is unblocked, pppd does not expect it */
+	sigfillset(&sigs);
+	sigprocmask(SIG_UNBLOCK, &sigs, NULL);
+	/* run pppd now */
 	execvp(pppd_argv[0], pppd_argv);
 	/* execvp() failed */
-	syslog(LOG_ERR, "CTRL (PPPD Launcher): Failed to launch PPP daemon.");
+	syslog(LOG_ERR, 
+	       "CTRL (PPPD Launcher): Failed to launch PPP daemon. %s",
+	       strerror(errno));
 }
+
