@@ -1,10 +1,10 @@
 /*
  * pptpd.c
  *
- * Grabs any command line argument and procecesses any further options in
+ * Grabs any command line argument and processes any further options in
  * the pptpd config file, before throwing over to pptpmanager.c.
  *
- * $Id: pptpd.c,v 1.2 1999/12/17 18:57:30 patl Exp $
+ * $Id: pptpd.c,v 1.11 2004/05/17 06:32:00 quozl Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -23,13 +23,10 @@
 #define _XOPEN_SOURCE 500	/* strdup() prototype */
 #endif
 
-#if HAVE_SYSLOG_H
-#include <syslog.h>
-#else
 #include "our_syslog.h"
-#endif
+#include "our_getopt.h"
 
-#include <ctype.h>
+#include <fcntl.h>
 #include <netdb.h>
 #include <signal.h>
 #include <stdio.h>
@@ -46,44 +43,68 @@
 #include "configfile.h"
 #include "defaults.h"
 #include "compat.h"
-#include "our_getopt.h"
 #include "pptpmanager.h"
 
+#ifdef CONFIG_NETtel
+#include <linux/ledman.h>
+#endif
+
 /* command line arg variables */
+char *ppp_binary = NULL;
 char *pppdoptstr = NULL;
 char *speedstr = NULL;
 char *bindaddr = NULL;
+#ifdef BCRELAY
+char *bcrelay = NULL;
+#endif
 int pptp_debug = 0;
+int pptp_noipparam = 0;
+int pptp_logwtmp = 0;
+
+#if defined(PPPD_IP_ALLOC)
+int pptp_stimeout = STIMEOUT_DEFAULT;
+#endif
 
 #if !defined(PPPD_IP_ALLOC)
 int maxConnections = MAX_CONNECTIONS;
 
 char localIP[MAX_CONNECTIONS][16];
 char remoteIP[MAX_CONNECTIONS][16];
-u_int32_t ipx_min = 0xffffffff;
 
 /* Local prototypes */
 static void processIPStr(int type, char *ipstr);
-static void ipxnumparse(char *netstring);
 #endif
 
 #ifndef HAVE_DAEMON
 static void my_daemon(int argc, char **argv);
 #endif
 
-static void log_pid(char *);
+static void log_pid(char *pid_file);
 static char *lookup(char *);
 
+#ifdef BCRELAY
+static void launch_bcrelay();
+static void killbcrelay(int sigraised);
+static pid_t bcrelayfork;
+#endif
+
 static void showusage(char *prog)
 {
-	printf("\nPoPToP v%s\n", VERSION);
+	printf("\nPoptop v%s\n", VERSION);
 	printf("The PPTP Server for Linux\n");
 	printf("Usage: pptpd [options], where options are:\n\n");
+#ifdef BCRELAY
+	printf(" [-b] [--bcrelay if]       Use broadcast relay for broadcasts comming from.\n");
+	printf("                           the specified interface (default is eth1).\n");
+#endif
 	printf(" [-c] [--conf file]        Specifies the config file to read default\n");
-	printf("                           settings from (default is /etc/pptpd.conf).\n");
+	printf("                           settings from (default is %s).\n", PPTPD_CONFIG_FILE_DEFAULT);
 	printf(" [-d] [--debug]            Turns on debugging (to syslog).\n");
+	printf(" [-e] [--ppp file]         Use alternate pppd binary, default %s.\n", PPP_BINARY);
 	printf(" [-f] [--fg]               Run in foreground.\n");
 	printf(" [-h] [--help]             Displays this help message.\n");
+	printf(" [-i] [--noipparam]        Suppress the passing of the client's IP address\n");
+	printf("                           to PPP, which is done by default otherwise.\n");
 	printf(" [-l] [--listen x.x.x.x]   Specifies IP of local interface to listen to.\n");
 #if !defined(BSDUSER_PPP)
 	printf(" [-o] [--option file]      Specifies the PPP options file to use\n");
@@ -95,18 +116,21 @@ static void showusage(char *prog)
 	printf(" [-s] [--speed baud]       Specifies the baud speed for the PPP daemon\n");
 	printf("                           (default is 115200).\n");
 #endif
+	printf(" [-t] [--stimeout seconds] Specifies the timeout for the first packet. This is a DOS protection\n");
+	printf("                           (default is 10).\n");
 	printf(" [-v] [--version]          Displays the PoPToP version number.\n");
+	printf(" [-w] [--logwtmp]          Update wtmp as users login.\n");
 
 	printf("\n\nLogs and debugging go to syslog as DAEMON.");
 
 	printf("\n\nCommand line options will override any default settings and any settings\n");
-	printf("specified in the config file (default config file: /etc/pptpd.conf).\n\n");
+	printf("specified in the config file (default config file: %s).\n\n", PPTPD_CONFIG_FILE_DEFAULT);
 }
 
 
 static void showversion()
 {
-	printf("PoPToP v%s\n", VERSION);
+	printf("Poptop v%s\n", VERSION);
 }
 
 int main(int argc, char **argv)
@@ -125,98 +149,123 @@ int main(int argc, char **argv)
 	char tmp[MAX_CONFIG_STRING_SIZE], *tmpstr;
 
 	/* open a connection to the syslog daemon */
-	openlog("pptpd", LOG_PID, LOG_PPTP);
+	openlog("pptpd", LOG_PID, LOG_DAEMON);
 
+	/* process command line options */
 	while (1) {
 		int option_index = 0;
+#ifdef BCRELAY
+		char *optstring = "b:c:de:fhil:o:p:s:t:vw";
+#else
+		char *optstring = "c:de:fhil:o:p:s:t:vw";
+#endif
 
 		static struct option long_options[] =
 		{
+#ifdef BCRELAY
+			{"bcrelay", 1, 0, 0},
+#endif
 			{"conf", 1, 0, 0},
 			{"debug", 0, 0, 0},
+			{"ppp", 1, 0, 0},
 			{"fg", 0, 0, 0},
 			{"help", 0, 0, 0},
+			{"noipparam", 0, 0, 0},
 			{"listen", 1, 0, 0},
 			{"option", 1, 0, 0},
 			{"pidfile", 1, 0, 0},
 			{"speed", 1, 0, 0},
+			{"stimeout", 1, 0, 0},
 			{"version", 0, 0, 0},
+			{"logwtmp", 0, 0, 0},
 			{0, 0, 0, 0}
 		};
 
-		c = getopt_long(argc, argv, "c:dfhl:o:p:s:v", long_options, &option_index);
+		c = getopt_long(argc, argv, optstring, long_options, &option_index);
 		if (c == -1)
 			break;
 		/* convert long options to short form */
 		if (c == 0)
-			c = "cdfhlopsv"[option_index];
+#ifdef BCRELAY
+			c = "bcdefhilopstvw"[option_index];
+#else
+			c = "cdefhilopstvw"[option_index];
+#endif
 		switch (c) {
-		case 'l':
+#ifdef BCRELAY
+		case 'b': /* --bcrelay */
+			if (bcrelay) free(bcrelay);
+			bcrelay = strdup(optarg);
+			break;
+#endif
+
+		case 'l': /* --listen */
 			tmpstr = lookup(optarg);
-			if(!tmpstr) {
-				fprintf(stderr, "Invalid listening address: %s!\n", optarg);
+			if (!tmpstr) {
 				syslog(LOG_ERR, "MGR: Invalid listening address: %s!", optarg);
 				return 1;
 			}
-			if(bindaddr)
-				free(bindaddr);
+			if (bindaddr) free(bindaddr);
 			bindaddr = strdup(tmpstr);
 			break;
 
-		case 'h':
+		case 'h': /* --help */
 			showusage(argv[0]);
 			return 0;
 
-		case 'd':
+		case 'i': /* --noipparam */
+			pptp_noipparam = TRUE;
+			break;
+
+		case 'e': /* --ppp */
+			if (ppp_binary) free(ppp_binary);
+			ppp_binary = strdup(optarg);
+			break;
+
+		case 'd': /* --debug */
 			pptp_debug = TRUE;
 			break;
 
-		case 'f':
+		case 'f': /* --fg */
 			foreground = TRUE;
 			break;
 
-		case 'v':
+		case 'v': /* --version */
 			showversion();
 			return 0;
 
-		case 'o':
-			{
-				FILE *f;
-				if (!(f = fopen(optarg, "r"))) {
-					fprintf(stderr, "PPP options file not found!\n");
-					syslog(LOG_ERR, "MGR: PPP options file not found!");
-					return 1;
-				}
-				fclose(f);
-				if(pppdoptstr)
-					free(pppdoptstr);
-				pppdoptstr = strdup(optarg);
-				break;
-			}
+		case 'w': /* --logwtmp */
+		        pptp_logwtmp = TRUE;
+			break;
 
-		case 'p':
-			if(pid_file)
-				free(pid_file);
+		case 'o': /* --option */
+			if (pppdoptstr) free(pppdoptstr);
+			pppdoptstr = strdup(optarg);
+			break;
+
+		case 'p': /* --pidfile */
+			if (pid_file) free(pid_file);
 			pid_file = strdup(optarg);
 			break;
 
-		case 's':
-			if(speedstr)
-				free(speedstr);
+		case 's': /* --speed */
+			if (speedstr) free(speedstr);
 			speedstr = strdup(optarg);
 			break;
 
-		case 'c':
+		case 't': /* --stimeout */
+			pptp_stimeout = atoi(optarg);
+			break;
+
+		case 'c': /* --conf */
 			{
 				FILE *f;
 				if (!(f = fopen(optarg, "r"))) {
-					fprintf(stderr, "Config file not found!\n");
 					syslog(LOG_ERR, "MGR: Config file not found!");
 					return 1;
 				}
 				fclose(f);
-				if(configFile)
-					free(configFile);
+				if(configFile) free(configFile);
 				configFile = strdup(optarg);
 				break;
 			}
@@ -238,10 +287,24 @@ int main(int argc, char **argv)
 	if (!pptp_debug && read_config_file(configFile, DEBUG_KEYWORD, tmp) > 0)
 		pptp_debug = TRUE;
 
+#ifdef BCRELAY
+	if (!bcrelay && read_config_file(configFile, BCRELAY_KEYWORD, tmp) > 0) 
+		bcrelay = strdup(tmp);
+#endif
+
+	if (!pptp_stimeout && read_config_file(configFile, STIMEOUT_KEYWORD, tmp) > 0) {
+		pptp_stimeout = atoi(tmp);
+		if (pptp_stimeout <= 0)
+			pptp_stimeout = STIMEOUT_DEFAULT;
+	}
+
+	if (!pptp_noipparam && read_config_file(configFile, NOIPPARAM_KEYWORD, tmp) > 0) {
+		pptp_noipparam = TRUE;
+	}
+
 	if (!bindaddr && read_config_file(configFile, LISTEN_KEYWORD, tmp) > 0) {
 		tmpstr = lookup(tmp);
 		if(!tmpstr) {
-			fprintf(stderr, "Invalid listening address: %s!\n", tmp);
 			syslog(LOG_ERR, "MGR: Invalid listening address: %s!", tmp);
 			return 1;
 		}
@@ -252,13 +315,17 @@ int main(int argc, char **argv)
 		speedstr = strdup(tmp);
 
 	if (!pppdoptstr && read_config_file(configFile, PPPD_OPTION_KEYWORD, tmp) > 0) {
-		if (!(fopen(tmp, "r"))) {
-			fprintf(stderr, "PPP options file not found!\n");
-			syslog(LOG_ERR, "MGR: PPP options file not found!");
-			return 1;
-		}
 		pppdoptstr = strdup(tmp);
 	}
+
+	if (!ppp_binary && read_config_file(configFile, PPP_BINARY_KEYWORD, tmp) > 0) {
+		ppp_binary = strdup(tmp);
+	}
+
+	if (!pptp_logwtmp && read_config_file(configFile, LOGWTMP_KEYWORD, tmp) > 0) {
+		pptp_logwtmp = TRUE;
+	}
+
 	if (!pid_file)
 		pid_file = strdup((read_config_file(configFile, PIDFILE_KEYWORD,
 					tmp) > 0) ? tmp : PIDFILE_DEFAULT);
@@ -279,23 +346,41 @@ int main(int argc, char **argv)
 		strlcpy(tmp, DEFAULT_LOCAL_IP_LIST, sizeof(tmp));
 	}
 	processIPStr(LOCAL, tmp);
-
-	/* parse config file IPX addresses */
-	if (read_config_file(configFile, IPXNETS_KEYWORD, tmp) > 0)
-		ipxnumparse(tmp);
 #endif
 
 	free(configFile);
 
+	/* if not yet set, adopt default PPP binary path */
+	if (!ppp_binary) ppp_binary = strdup(PPP_BINARY);
+	/* check that the PPP binary is executable */
+	if (access(ppp_binary, X_OK) < 0) {
+		syslog(LOG_ERR, "MGR: PPP binary %s not executable",
+		       ppp_binary);
+		return 1;
+	}
+	/* check that the PPP options file is readable */
+	if (pppdoptstr && access(pppdoptstr, R_OK) < 0) {
+		syslog(LOG_ERR, "MGR: PPP options file %s not readable",
+		       pppdoptstr);
+		return 1;
+	}
+#ifdef BCRELAY
+	/* check that the bcrelay binary is executable */
+	if (bcrelay && access(BCRELAY_BIN, X_OK) < 0) {
+		syslog(LOG_ERR, "MGR: bcrelay binary %s not executable", 
+		       BCRELAY_BIN);
+		return 1;
+	}
+#endif
+
 	if (!foreground) {
 #if HAVE_DAEMON
 		closelog();
 		freopen("/dev/null", "r", stdin);
-		/* set noclose, we want stdout/stderr still attached if we can */
-		daemon(0, 1);
+		daemon(0, 0);
 		/* returns to child only */
 		/* pid will have changed */
-		openlog("pptpd", LOG_PID, LOG_PPTP);
+		openlog("pptpd", LOG_PID, LOG_DAEMON);
 #else	/* !HAVE_DAEMON */
 		my_daemon(argc, argv);
 		/* returns to child if !HAVE_FORK
@@ -304,33 +389,71 @@ int main(int argc, char **argv)
 #endif
 	}
 
+#ifdef BCRELAY
+      if (bcrelay) {
+             syslog(LOG_DEBUG, "CTRL: BCrelay incoming interface is %s", bcrelay);
+             /* Launch BCrelay  */
+#ifndef HAVE_FORK
+             switch(bcrelayfork = vfork()){
+#else
+             switch(bcrelayfork = fork()){
+#endif
+             case -1:        /* fork() error */
+                   syslog(LOG_ERR, "CTRL: Error forking to exec bcrelay");
+                   _exit(1);
+
+             case 0:         /* child */
+                   syslog(LOG_DEBUG, "CTRL (BCrelay Launcher): Launching BCrelay with pid %i", bcrelayfork);
+                   launch_bcrelay();
+                   syslog(LOG_ERR, "CTRL (BCrelay Launcher): Failed to launch BCrelay.");
+                   _exit(1);
+             }
+       } /* End bcrelay */
+#endif
+
+#ifdef CONFIG_NETtel
+	/* turn the NETtel VPN LED on */
+	ledman_cmd(LEDMAN_CMD_ON, LEDMAN_VPN);
+#endif
 	/* after we have our final pid... */
 	log_pid(pid_file);
-	free(pid_file);
+
+#ifdef BCRELAY
+        /* be ready for a grisly death */
+        signal(SIGTERM, &killbcrelay);
+#endif
 
 	pptp_manager(argc, argv);
-	return 1;
+
+        return 1;
+
 }
 
+#ifdef BCRELAY
+static void killbcrelay(int sigraised) {
+        if (sigraised) {
+                syslog(LOG_DEBUG, "CTRL: Closing child BCrelay with pid %i", bcrelayfork);
+                if (bcrelayfork > 0)
+                kill(bcrelayfork, SIGQUIT);
+		exit(0);
+        }
+}
+#endif
+
 static void log_pid(char *pid_file) {
-	FILE *f;
-	int i;
+        FILE    *f;
+        pid_t   pid;
 
-	i = umask(022);
-	f = fopen(pid_file, "w");
-	umask(i);
-	if(!f)
-		return;
-	fprintf(f, "%d\n", getpid());
-	fclose(f);
+        pid = getpid();
+        if ((f = fopen(pid_file, "w")) == NULL) {
+                syslog(LOG_ERR, "PPTPD: failed to open(%s), errno=%d\n",
+                        pid_file, errno);
+                return;
+        }
+        fprintf(f, "%d\n", pid);
+        fclose(f);
 }
 
-#if HAVE_SETSID
-#define SETSIDPGRP setsid
-#else
-#define SETSIDPGRP setpgrp
-#endif
-
 #ifndef HAVE_DAEMON
 static void my_daemon(int argc, char **argv)
 {
@@ -338,25 +461,27 @@ static void my_daemon(int argc, char **argv)
 	/* need to use vfork - eg, uClinux */
 	char **new_argv;
 	int pid;
+	extern char **environ;
+	int fdr;
+
+	new_argv = malloc((argc + 2) * sizeof(char **));
+	fdr = open("/dev/null", O_RDONLY);
 	syslog(LOG_INFO, "MGR: Option parse OK, re-execing as daemon");
-	fprintf(stderr, "pptpd: option parse OK, re-execing as daemon\n");
 	fflush(stderr);
 	if ((pid = vfork()) == 0) {
-		freopen("/dev/null", "r", stdin);
+		if (fdr != 0) { dup2(fdr, 0); close(fdr); }
 		SETSIDPGRP();
 		chdir("/");
 		umask(0);
-		new_argv = malloc((argc + 2) * sizeof(char **));
 		memcpy(new_argv + 1, argv, (argc + 1) * sizeof(char **));
 		new_argv[0] = PPTPD_BIN;
 		new_argv[1] = "-f";
-		execvp(PPTPD_BIN, new_argv);
-		perror("execvp");
-		exit(1);
+		execve(PPTPD_BIN, new_argv, environ);
+		_exit(1);
 	} else if (pid > 0) {
 		exit(0);
 	} else {
-		perror("vfork");
+		syslog_perror("vfork");
 		exit(1);
 	}
 #else
@@ -364,7 +489,7 @@ static void my_daemon(int argc, char **argv)
 
 	closelog();
 	if ((pid = fork()) < 0) {
-		perror("fork");
+		syslog_perror("fork");
 		exit(1);
 	} else if (pid)
 		exit(0);
@@ -373,7 +498,7 @@ static void my_daemon(int argc, char **argv)
 	chdir("/");
 	umask(0);
 	/* pid will have changed */
-	openlog("pptpd", LOG_PID, LOG_PPTP);
+	openlog("pptpd", LOG_PID, LOG_DAEMON);
 #endif
 }
 #endif
@@ -493,7 +618,6 @@ static void processIPStr(int type, char *ipstr)
 			 */
 			if ((tmpstr7 = lookup(tmpstr2)) == NULL) {
 				syslog(LOG_ERR, "MGR: Bad IP address (%s) in config file!", tmpstr2);
-				fprintf(stderr, "Bad IP address (%s) in config file!\n", tmpstr2);
 				exit(1);
 			}
 			if (num == MAX_CONNECTIONS) {
@@ -610,7 +734,6 @@ static void processIPStr(int type, char *ipstr)
 				/* Check if the ip address is valid */
 				if ((tmpstr7 = validip(tmpstr5)) == NULL) {
 					syslog(LOG_ERR, "MGR: Bad IP address (%s) in config file!", tmpstr5);
-					fprintf(stderr, "Bad IP address (%s) in config file!\n", tmpstr5);
 					exit(1);
 				}
 				if (num == MAX_CONNECTIONS) {
@@ -637,26 +760,32 @@ static void processIPStr(int type, char *ipstr)
 	} else if (maxConnections > num)
 		maxConnections = num;
 }
+#endif
 
-static void ipxbadparse() {
-	syslog(LOG_ERR, "MGR: IPX - can't parse addresses or invalid range!");
-	fprintf(stderr, "Can't parse IPX addresses or invalid range!\n");
-	exit(1);
-}
+#ifdef BCRELAY
+/* launch_bcrelay
+ * Launches broadcast relay. Broadcast relay is responsible for relaying broadcasts to the clients
+ * retn: 0 on success, -1 on failure.
+ */
+static void launch_bcrelay() {
+  char *bcrelay_argv[6];
+  int an = 0;
 
-static void ipxnumparse(char *netstring) {
-	u_int32_t ipx_max;
+      if (bcrelay) {
+           syslog(LOG_DEBUG, "MGR: BCrelay incoming interface is %s", bcrelay);
+           syslog(LOG_DEBUG, "MGR: BCrelay outgoing interface is regexp ppp[0-9].*");
 
-	ipx_min = strtoul(netstring, &netstring, 16);
-	if(*netstring != '-')
-		ipxbadparse();
-	netstring++;
-	if(!*netstring)
-		ipxbadparse();
-	ipx_max = strtoul(netstring, &netstring, 16);
-	if(ipx_max < ipx_min)
-		ipxbadparse();
-	if(maxConnections > ipx_max - ipx_min)
-		maxConnections = ipx_max - ipx_min;
+	   bcrelay_argv[an++] = BCRELAY_BIN;
+	   bcrelay_argv[an++] = "-i";
+	   bcrelay_argv[an++] = bcrelay;
+	   bcrelay_argv[an++] = "-o";
+	   bcrelay_argv[an++] = "ppp[0-9].*";
+           if (!pptp_debug) {
+	         bcrelay_argv[an++] = "-n";
+           }
+	   bcrelay_argv[an++] = NULL;
+
+           execvp(bcrelay_argv[0], bcrelay_argv);
+      }
 }
 #endif
