@@ -3,7 +3,7 @@
  *
  * Manages the PoPToP sessions.
  *
- * $Id: pptpmanager.c,v 1.1.1.1 1999/12/17 16:30:14 patl Exp $
+ * $Id: pptpmanager.c,v 1.7 2004/04/28 11:36:07 quozl Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -14,11 +14,7 @@
 #define _GNU_SOURCE 1		/* broken arpa/inet.h */
 #endif
 
-#if HAVE_SYSLOG_H
-#include <syslog.h>
-#else
 #include "our_syslog.h"
-#endif
 
 #include <errno.h>
 #include <netdb.h>
@@ -33,6 +29,7 @@
 #include <sys/un.h>
 #include <sys/wait.h>
 #include <unistd.h>
+#include <time.h>
 #include <sys/time.h>
 #include <fcntl.h>
 
@@ -45,6 +42,10 @@ int allow_severity = LOG_WARNING;
 int deny_severity = LOG_WARNING;
 #endif
 
+#ifdef __UCLIBC__
+#define socklen_t int
+#endif
+
 #include "configfile.h"
 #include "defaults.h"
 #include "pptpctrl.h"
@@ -53,15 +54,17 @@ int deny_severity = LOG_WARNING;
 #include "compat.h"
 
 /* command line arg variables */
+extern char *ppp_binary;
 extern char *pppdoptstr;
 extern char *speedstr;
 extern char *bindaddr;
 extern int pptp_debug;
+extern int pptp_noipparam;
+extern int pptp_logwtmp;
 
 #if !defined(PPPD_IP_ALLOC)
 extern char localIP[MAX_CONNECTIONS][16];
 extern char remoteIP[MAX_CONNECTIONS][16];
-extern u_int32_t ipx_min;
 
 /* for now, this contains all relavant info about a call
  * we are assuming that the remote and local IP's never change
@@ -73,6 +76,9 @@ struct callArray {
 	char pppLocal[16];
 };
 
+/* option for timeout on starting ctrl connection */
+int pptp_stimeout = STIMEOUT_DEFAULT;
+
 /* global for signal handler */
 static struct callArray clientArray[MAX_CONNECTIONS];
 
@@ -84,6 +90,9 @@ extern int maxConnections;
 static void connectCall(int clientSocket, int clientNumber);
 static int createHostSocket(int *hostSocket);
 
+/* this end's call identifier */
+uint16_t unique_call_id = 0;
+
 static void pptp_reap_child(int sig)
 {
 	int chld, status;
@@ -111,20 +120,32 @@ static void pptp_reap_child(int sig)
 int pptp_manager(int argc, char **argv)
 {
 	/* misc ints */
-	int loop, firstOpen, ctrl_pid;
+#if !defined(PPPD_IP_ALLOC)
+	int loop;
+#endif
+	int ctrl_pid;
 	socklen_t addrsize;
 
 	/* for host stuff */
 	int hostSocket;
 	fd_set connSet;
+	sigset_t sigchld;
+	struct sigaction sa;
 
-	signal(SIGCHLD, pptp_reap_child);
+	sigemptyset(&sigchld);
+	sigaddset(&sigchld, SIGCHLD);
+
+	memset(&sa, 0, sizeof(sa));
+	sa.sa_handler = pptp_reap_child;
+	sigaction(SIGCHLD, &sa, NULL);
 
 	/* openlog() not required, done in pptpd.c */
 
 	syslog(LOG_INFO, "MGR: Manager process started");
 
 #if !defined(PPPD_IP_ALLOC)
+	syslog(LOG_INFO, "MGR: Maximum of %d connections available", maxConnections);
+
 	for (loop = 0; loop < maxConnections; loop++) {
 		clientArray[loop].pid = 0;
 	}
@@ -133,26 +154,24 @@ int pptp_manager(int argc, char **argv)
 	/* Connect the host socket and activate it for listening */
 	if (createHostSocket(&hostSocket) < 0) {
 		syslog(LOG_ERR, "MGR: Couldn't create host socket");
-		perror("createHostSocket");
+		syslog_perror("createHostSocket");
 		exit(-1);
 	}
 
 	FD_ZERO(&connSet);
 
 	while (1) {
-		firstOpen = -1;
-		loop = 0;
-
 #if !defined(PPPD_IP_ALLOC)
-		while ((loop < maxConnections) && (firstOpen == -1)) {
-			if ((clientArray[loop].pid == 0) && (firstOpen == -1))
+		int firstOpen = -1;
+
+		for (loop = 0; loop < maxConnections; loop++)
+			if (clientArray[loop].pid == 0) {
 				firstOpen = loop;
-			loop++;
-		}
+				break;
+			}
 
 		if (firstOpen == -1) {
 			syslog(LOG_ERR, "MGR: No free connection slots or IPs - no more clients can connect!");
-			fprintf(stderr, "No free connection slots or IPs available - no more clients can connect!\n");
 			FD_CLR(hostSocket, &connSet);
 		} else {
 			FD_SET(hostSocket, &connSet);
@@ -166,7 +185,7 @@ int pptp_manager(int argc, char **argv)
 		 */
 		if (select(hostSocket + 1, &connSet, NULL, NULL, NULL) < 0 && errno != EINTR) {
 			syslog(LOG_ERR, "MGR: Error with manager select()!");
-			perror("select");
+			syslog_perror("select");
 			exit(-1);
 		}
 
@@ -199,9 +218,72 @@ int pptp_manager(int argc, char **argv)
 			if (clientSocket == -1) {
 				/* accept failed, but life goes on... */
 				syslog(LOG_ERR, "MGR: accept() failed");
-				perror("accept");
+				syslog_perror("accept");
 			} else if (clientSocket != 0) {
+				fd_set rfds;
+				struct timeval tv;
+				struct pptp_header ph;
 
+				/*
+				 * DOS protection: get a peek at the first packet
+				 * and do some checks on it before we continue.
+				 * A 10 second timeout on the first packet seems reasonable
+				 * to me,  if anything looks sus,  throw it away.
+				 */
+
+				FD_ZERO(&rfds);
+				FD_SET(clientSocket, &rfds);
+				tv.tv_sec = pptp_stimeout;
+				tv.tv_usec = 0;
+				if (select(clientSocket + 1, &rfds, NULL, NULL, &tv) <= 0) {
+					syslog(LOG_ERR, "MGR: dropped slow initial connection");
+					close(clientSocket);
+					continue;
+				}
+
+				if (recv(clientSocket, &ph, sizeof(ph), MSG_PEEK) !=
+						sizeof(ph)) {
+					syslog(LOG_ERR, "MGR: dropped small initial connection");
+					close(clientSocket);
+					continue;
+				}
+
+				ph.length = htons(ph.length);
+				ph.pptp_type = ntohs(ph.pptp_type);
+				ph.magic = ntohl(ph.magic);
+				ph.ctrl_type = ntohs(ph.ctrl_type);
+
+				if (ph.length <= 0 || ph.length > PPTP_MAX_CTRL_PCKT_SIZE) {
+					syslog(LOG_WARNING, "MGR: initial packet length %d outside "
+							"(0 - %d)",  ph.length, PPTP_MAX_CTRL_PCKT_SIZE);
+					goto dos_exit;
+				}
+
+				if (ph.magic != PPTP_MAGIC_COOKIE) {
+					syslog(LOG_WARNING, "MGR: initial packet bad magic");
+					goto dos_exit;
+				}
+
+				if (ph.pptp_type != PPTP_CTRL_MESSAGE) {
+					syslog(LOG_WARNING, "MGR: initial packet has bad type");
+					goto dos_exit;
+				}
+
+				if (ph.ctrl_type != START_CTRL_CONN_RQST) {
+					syslog(LOG_WARNING, "MGR: initial packet has bad ctrl type "
+							"0x%x", ph.ctrl_type);
+		dos_exit:
+					close(clientSocket);
+					continue;
+				}
+
+				/* block SIGCHLD until parent registers
+				 * child in the connection list, else we
+				 * have a race condition with the handler.
+				 */
+				if (sigprocmask(SIG_BLOCK, &sigchld, NULL) !=0) {
+					syslog_perror("sigprocmask");
+				}
 #ifndef HAVE_FORK
 				switch (ctrl_pid = vfork()) {
 #else
@@ -221,13 +303,17 @@ int pptp_manager(int argc, char **argv)
 #else
 					connectCall(clientSocket, 0);
 #endif
-					exit(1);
+					_exit(1);
 					/* NORETURN */
 				default:	/* parent */
 					close(clientSocket);
+					unique_call_id += MAX_CALLS_PER_TCP_LINK;
 #if !defined(PPPD_IP_ALLOC)
 					clientArray[firstOpen].pid = ctrl_pid;
 #endif
+					if (sigprocmask(SIG_UNBLOCK, &sigchld, NULL) !=0) {
+						syslog_perror("sigprocmask");
+					}
 					break;
 				}
 			}
@@ -285,8 +371,8 @@ static int createHostSocket(int *hostSocket)
 	if (bind(*hostSocket, (struct sockaddr *) &address, sizeof(address)) < 0)
 		return -3;
 
-	/* listen to as many connections as we can, but only accept() when a slot is free */
-	if (listen(*hostSocket, SOMAXCONN) < 0)
+	/* minimal backlog to avoid DoS */
+	if (listen(*hostSocket, 3) < 0)
 		return -4;
 
 	return 0;
@@ -301,14 +387,20 @@ static int createHostSocket(int *hostSocket)
 static void connectCall(int clientSocket, int clientNumber)
 {
 
+#define NUM2ARRAY(array, num) snprintf(array, sizeof(array), "%d", num)
+	
 	char *ctrl_argv[16];	/* arguments for launching 'pptpctrl' binary */
 
 	int pptpctrl_argc = 0;	/* count the number of arguments sent to pptpctrl */
 
 	/* lame strings to hold passed args. */
-	char ctrl_debug[16];
-	char pppdoptfile_argv[16];
-	char speedgiven_argv[16];
+	char ctrl_debug[2];
+	char ctrl_noipparam[2];
+	char pppdoptfile_argv[2];
+	char speedgiven_argv[2];
+	extern char **environ;
+
+	char callid_argv[16];
 
 	/*
 	 * Launch the CTRL manager binary; we send it some information such as
@@ -318,22 +410,31 @@ static void connectCall(int clientSocket, int clientNumber)
 	ctrl_argv[pptpctrl_argc++] = PPTP_CTRL_BIN "                                ";
 
 	/* Pass socket as stdin */
-	dup2(clientSocket, 0);
-	close(clientSocket);
+	if (clientSocket != 0) {
+		dup2(clientSocket, 0);
+		close(clientSocket);
+	}
 
 	/* get argv set up */
-	sprintf(ctrl_debug, "%d", pptp_debug);
+	NUM2ARRAY(ctrl_debug, pptp_debug ? 1 : 0);
+	ctrl_debug[1] = '\0';
 	ctrl_argv[pptpctrl_argc++] = ctrl_debug;
 
+	NUM2ARRAY(ctrl_noipparam, pptp_noipparam ? 1 : 0);
+	ctrl_noipparam[1] = '\0';
+	ctrl_argv[pptpctrl_argc++] = ctrl_noipparam;
+
 	/* optionfile = TRUE or FALSE; so the CTRL manager knows whether to load a non-standard options file */
-	sprintf(pppdoptfile_argv, "%d", pppdoptstr ? 1 : 0);
+	NUM2ARRAY(pppdoptfile_argv, pppdoptstr ? 1 : 0);
+	pppdoptfile_argv[1] = '\0';
 	ctrl_argv[pptpctrl_argc++] = pppdoptfile_argv;
 	if (pppdoptstr) {
 		/* send the option filename so the CTRL manager can launch pppd with this alternate file */
 		ctrl_argv[pptpctrl_argc++] = pppdoptstr;
 	}
 	/* tell the ctrl manager whether we were given a speed */
-	sprintf(speedgiven_argv, "%d", speedstr ? 1 : 0);
+	NUM2ARRAY(speedgiven_argv, speedstr ? 1 : 0);
+	speedgiven_argv[1] = '\0';
 	ctrl_argv[pptpctrl_argc++] = speedgiven_argv;
 	if (speedstr) {
 		/* send the CTRL manager the speed of the connection so it can fire pppd at that speed */
@@ -341,29 +442,35 @@ static void connectCall(int clientSocket, int clientNumber)
 	}
 #if PPPD_IP_ALLOC
 	/* no local or remote address to specify */
-	ctrl_argv[pptpctrl_argc++] = strdup("0");
-	ctrl_argv[pptpctrl_argc++] = strdup("0");
-	ctrl_argv[pptpctrl_argc++] = strdup("0");
+	ctrl_argv[pptpctrl_argc++] = "0";
+	ctrl_argv[pptpctrl_argc++] = "0";
 #else
 	/* specify local & remote addresses for this call */
-	ctrl_argv[pptpctrl_argc++] = strdup("1");
+	ctrl_argv[pptpctrl_argc++] = "1";
 	ctrl_argv[pptpctrl_argc++] = localIP[clientNumber];
-	ctrl_argv[pptpctrl_argc++] = strdup("1");
+	ctrl_argv[pptpctrl_argc++] = "1";
 	ctrl_argv[pptpctrl_argc++] = remoteIP[clientNumber];
-	if(ipx_min != 0xffffffff) {
-		ctrl_argv[pptpctrl_argc++] = strdup("1");
-		sprintf((ctrl_argv[pptpctrl_argc++] = malloc(9)),
-			"%08x", ipx_min + clientNumber);
-	} else
-		ctrl_argv[pptpctrl_argc++] = strdup("0");
 #endif
+
+	/* our call id to be included in GRE packets the client
+	 * will send to us */
+	NUM2ARRAY(callid_argv, unique_call_id);
+	ctrl_argv[pptpctrl_argc++] = callid_argv;
+
+	/* pass path to ppp binary */
+	ctrl_argv[pptpctrl_argc++] = ppp_binary;
+
+	/* pass logwtmp flag */
+	ctrl_argv[pptpctrl_argc++] = pptp_logwtmp ? "1" : "0";
+
+	/* note: update pptpctrl.8 if the argument list format is changed */
+
 	/* terminate argv array with a NULL */
 	ctrl_argv[pptpctrl_argc] = NULL;
 	pptpctrl_argc++;
 
 	/* ok, args are setup: invoke the call handler */
-	execvp(PPTP_CTRL_BIN, ctrl_argv);
+	execve(PPTP_CTRL_BIN, ctrl_argv, environ);
 	syslog(LOG_ERR, "MGR: Failed to exec " PPTP_CTRL_BIN "!");
-	perror("execvp " PPTP_CTRL_BIN);
-	exit(1);
+	_exit(1);
 }
